/*
** Module   : SocketListenerImpl.java
** Abstract : A class for implementing the SERVER-SOCKET method and attributes
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------------------Description---------------------------------
** 001 CS  20130204 Created initial version.
** 002 CA  20130221 Added Deletable support. setType was renamed to setResourceType.
** 003 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 004 CS  20130404 Implement runtime support
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import java.nio.channels.FileChannel.*;


import com.goldencode.p2j.net.*;

/**
 * A class for implementing the SERVER-SOCKET method and attributes.
 * 
 * TODO: implement resource delete.
 * TODO: what happens if the socket is deleted without a prior disconnect?
 */ 
public class SocketListenerImpl 
extends SensitiveResource
implements SocketListener,
           Sensitive
{   
   /** default keyalias*/
   private String keyalias = "default_server";
   
   /** Boolean flag to keep listening*/
   private boolean keepListening = true;
   
   /** The server socket id*/
   private long serverSocketId = -1;
   
   /** Client side interface for handling sockets */
   private LowLevelSocketListener clientSocketListener;

   private String connectProcedure;

   private handle contextHandle;
   
   /**
    * Default constructor which initializes SERVER-SOCKET specific settings.
    */
   public SocketListenerImpl()
   {
      // TODO implement constructor logic EX:
      clientSocketListener = (LowLevelSocketListener) RemoteObject.obtainNetworkInstance(LowLevelSocketListener.class);
      setResourceType("SERVER-SOCKET");
      setUnknown(false);
   }
   
   /**
    * Constructor with PREV-SIBLING and NEXT-SIBLING as attributes.
    * 
    * @param prevSibling
    *        The previously created Socket handle
    * @param nextSibling
    *        The following Socket handle
    */
   public SocketListenerImpl(handle prevSibling, handle nextSibling)
   {
      this();
      setPrevSibling(prevSibling);
      setNextSibling(nextSibling);
   }

   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      // The server socket becomes valid from the moment the CREATE SERVER-SOCKET instruction 
      // is called, this is equivalent to class being instantiated, disconnecting does not make 
      // the server socket invalid
      return true;
   }
   
   /**
    * Java implementation for the ENABLE-CONNECTION()  SERVER-SOCKET method, 
    * this will attempt to start listening  on the port specified by the 
    * option argument, and will return <code>true</code> if operation 
    * succeeded, or <code>false</code> otherwise.
    * <p>
    * The following options can be specified as valid ENABLE-CONNECTIONS() 
    * options:
    * <ul>
    *    <li><b>-S port</b> The port to which to connect.
    *    <li><b>-ssl</b> If specified the connection will be made using ssl.
    *    <li><b>-pf config-file-name</b> Specifies the file from which options
    *        will be taken.
    *    <li><b>-qsize maxconn</b> The maximum number of connections to handle 
    *        before refusing extra incoming connections.
    *    <li><b>-keyalias alias</b> The alias for the key and digital 
    *        certificates, default is : default_server.
    *   
    *    <li><b>-keyaliaspasswd pwd</b> The password for the key and digital
    *        certificate. To use always with -keyalias option.
    *    <li><b>-nossesioncache</b> Disables SSL client session cache.
    *    <li><b>-sessiontimeout time</b> The number of seconds before the SSL
    *        client request is timeout, default value is 180..
    * </ul>
    * 
    * @param  option
    *         The set of options that define the enable-connections method.
    * 
    * @return <code>true</code> if operation succeeded, or <code>false</code>
    *         otherwise
    */
   public logical enableConnections(String option)
   {
      return enableConnections(new character(option));
   }
   
   /**
    * Java implementation for the ENABLE-CONNECTION()  SERVER-SOCKET method, 
    * this will attempt to start listening  on the port specified by the 
    * option argument, and will return <code>true</code> if operation 
    * succeeded, or <code>false</code> otherwise.
    * <p>
    * The following options can be specified as valid ENABLE-CONNECTIONS() 
    * options:
    * <ul>
    *    <li><b>-S port</b> The port to which to connect.
    *    <li><b>-ssl</b> If specified the connection will be made using ssl.
    *    <li><b>-pf config-file-name</b> Specifies the file from which options
    *        will be taken.
    *    <li><b>-qsize maxconn</b> The maximum number of connections to handle 
    *        before refusing extra incoming connections.
    *    <li><b>-keyalias alias</b> The alias for the key and digital 
    *        certificates, default is : default_server.
    *   
    *    <li><b>-keyaliaspasswd pwd</b> The password for the key and digital
    *        certificate. To use always with -keyalias option.
    *    <li><b>-nossesioncache</b> Disables SSL client session cache.
    *    <li><b>-sessiontimeout time</b> The number of seconds before the SSL
    *        client request is timeout, default value is 180.
    * </ul>
    * 
    * @param  option
    *         The set of options that define the enable-connections method.
    * 
    * @return <code>true</code> if operation succeeded, or <code>false</code>
    *         otherwise.
    */
   public logical enableConnections(character option)
   {
      integer errNo = new integer();
      integer port = new integer();
      character keyAlias = new character();
      // qsize -1 value will be handles as default and not set.
      integer qsize = new integer(-1);
      integer timeout = new integer(180000);
      character fileOpts = new character(option);
      character errMsg = new character();
      character keyAliasPwd = new character();
      logical ssl = new logical(false);
      logical noSessionCache = new logical(false);
      
      getParams(fileOpts, errMsg, errNo, port, qsize, timeout, keyAlias, 
                keyAliasPwd, ssl, noSessionCache);
      
      if (!errMsg.isUnknown())
      {
         ErrorManager.recordOrThrowError(errNo.intValue(), errMsg.toStringMessage());
         ErrorManager.recordOrThrowError(5509, "Unable to process parameters.");
         ErrorManager.recordOrThrowError(5510, "Invalid parameter string.");
         return new logical(false);
      }
      
      if (!ssl.booleanValue())
      {

         serverSocketId = clientSocketListener.enableConnections(port.intValue(),
                                                                 timeout.intValue(),
                                                                 qsize.intValue());
      }
      else
      {

         char[] keyPassword = null;
         char[] certPwd = null;
         byte[] keyInput = null;
         int size = 0;

         FileInputStream fis = null;
         FileChannel channel = null;

         if (!keyAlias.isUnknown())
         {
            // TODO Check if this will match requirements
            keyalias = keyAlias.toStringMessage();

         }
         File keyStoreFile = new File(keyalias);
         if (keyStoreFile.exists())
         {
            try
            {
               fis = new FileInputStream(keyStoreFile);
               channel = fis.getChannel();
               size = (int) channel.size();
               MappedByteBuffer buf = channel.map(MapMode.READ_ONLY, 0, size);
               keyInput = new byte[size];
               buf.get(keyInput);
            }
            catch (IOException e)
            {
               // TODO Auto-generated catch block
               e.printStackTrace();
            }
            finally
            {
               try
               {
                  if (fis != null)
                  {
                     fis.close();
                  }
                  if (channel != null)
                  {
                     channel.close();
                  }
               }
               catch (IOException e)
               {
                  // nothing else can be done
               }
            }
         }
         else
         {
            // Error here
         }
         if (!keyAliasPwd.isUnknown())
         {
            keyPassword = keyAliasPwd.toStringMessage().toCharArray();
         }
         serverSocketId = clientSocketListener.enableSSLConnections(port.intValue(),
                                                                    timeout.intValue(),
                                                                    qsize.intValue(),
                                                                    keyInput,
                                                                    keyPassword,
                                                                    certPwd);

      }
      
      return new logical(true);
   }

   /**
    * Java implementation for the DISABLE-CONNECTIONS() SERVER-SOCKET method, 
    * this will stop listening and accepting any new connections, but without
    * interrupting the existing ones that haven't finished, and returns 
    * <code>true</code> if succeeded or  <code>false</code> otherwise.
    * 
    * @return <code>true</code> if the operation succeeded or <code>false</code> 
    *         otherwise.
    */
   public logical disableConnections()
   {
      return new logical(true);
   }
   
   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method, 
    * this will set the procedure to execute when a CONNECT event is 
    * received.
    * <p>
    * The set-connect-procedure can also contain a handle to a higher 
    * level procedure which contains the read response procedure (to resolve 
    * any procedure context issues).
    * 
    * @param  procedureName
    *         The name of the procedure we want to set for read response.
    *         
    * @return <code>true</code> if the operation was successful or 
    *         <code>false</code> otherwise.
    */
   public logical setConnectProcedure(String procedureName)
   {
      return setConnectProcedure(new character(procedureName));
   }

   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method, 
    * this will set the procedure to execute when a CONNECT event is 
    * received.
    * <p>
    * The set-connect-procedure can also contain a handle to a higher 
    * level procedure which contains the read response procedure (to resolve 
    * any procedure context issues).
    * 
    * @param  procedureName
    *         The name of the procedure we want to set for read response.
    *         
    * @return <code>true</code> if the operation was successful or 
    *         <code>false</code> otherwise.
    */
   public logical setConnectProcedure(character procedureName)
   {
      return setConnectProcedure(procedureName, null);
   }

   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method, 
    * this will set the procedure to execute when a CONNECT event is 
    * received.
    * <p>
    * The set-connect-procedure can also contain a handle to a higher 
    * level procedure which contains the read response procedure (to resolve 
    * any procedure context issues).
    * 
    * @param  procedureName
    *         The name of the procedure we want to set for read response.
    * @param  procHandle
    *         The handle for the procedure containing the procedure we want
    *         to set as connect procedure - optional.
    *         
    * @return <code>true</code> if the operation was successful or 
    *         <code>false</code> otherwise.
    */
   public logical setConnectProcedure(String procedureName, handle procHandle)
   {
      return setConnectProcedure(new character(procedureName), procHandle);
   }

   /**
    * Java implementation for the SET-CONNECT-PROCEDURE() SERVER-SOCKET method, 
    * this will set the procedure to execute when a CONNECT event is 
    * received.
    * <p>
    * The set-connect-procedure can also contain a handle to a higher 
    * level procedure which contains the read response procedure (to resolve 
    * any procedure context issues).
    * 
    * @param  procedureName
    *         The name of the procedure we want to set for read response.
    * @param  procHandle
    *         The handle for the procedure containing the procedure we want
    *         to set as connect procedure - optional.
    *         
    * @return <code>true</code> if the operation was successful or 
    *         <code>false</code> otherwise.
    */
   public logical setConnectProcedure(character procedureName,
                                      handle procHandle)
   {
      if (!procedureName.isUnknown())
      {
         connectProcedure = procedureName.toStringMessage();
      }
      contextHandle = procHandle;
      return new logical(true);
   }

   /**
    * Delete the resource.
    */
   @Override
   protected void resourceDelete()
   {
      // TODO Auto-generated method stub
   }
   /**
    * Start receiving connections until told to end
    */
   private void startListening()
   {
      Thread t = new Thread()
      {
         public void run()
         {
            
            while (keepListening)
            {
               // TODO placeholder until a more clear implementation
               long socket = clientSocketListener.accept(serverSocketId);
               // we can create a SocketImpl if we have the id?
            }
         }
      };
      t.start();
   }

   /**
    * Utility method for parsing the enableConnections parameters
    * @param options
    *        The string to parse.
    * @param errMsg
    *        The error message to set.
    * @param errNo
    *        The error number to set.
    * @param port
    *        The port which will be parse.
    * @param keyAlias
    *        The alias for the key and digital certificates to be parsed.
    * @param keyAliasPwd
    *        The password for the key and digital certificate to parse.
    * @param ssl
    *        True if -ssl parameter will be parsed.
    * @param noSessionCache
    *        The -nosessioncache parameter that will be parsed.
    */
   private void getParams(character options,
                          character errMsg,
                          integer   errNo,
                          integer   port,
                          integer   qsize, 
                          integer   timeout,
                          character keyAlias,
                          character keyAliasPwd,
                          logical   ssl,
                          logical   noSessionCache)
   {
    
      String strParams = options.toStringMessage();
      
      if (!strParams.toUpperCase().contains("-S") && !strParams.toUpperCase().contains("-PF"))
      {
         errMsg.assign("The -S parameter must be specified to connect");
         errNo.assign(5488);
      }
      
      String[] tokens = strParams.split(" ");
      
      for (int i = 0; i < tokens.length; i++)
      {
         if (!tokens[i].startsWith("-"))
         {
            errMsg.assign("Could not recognize arguments: " + tokens[i]);
            errNo.assign(301);
         }
         else
         {
            // if the option is the last element error
            if (i + 1 > tokens.length)
            {
               // Progress 4GL doesn't check if the argument is a valid accepted argument
               // this is just replicating the behavior
               errMsg.assign("You have not supplied a parameter for argument: " + tokens[i]);
               errNo .assign(1403);
               break;
            }
            if (tokens[i].equalsIgnoreCase("-S"))
            {
               try
               {
                  port.assign(Integer.parseInt(tokens[i+1]));
                  i++;
               }
               catch(NumberFormatException e)
               {
                  errMsg.assign("Unknown service " + tokens[i+1] + " transport TCP");
                  errNo.assign(5484);
                  break;
               }               
            }
            else if (tokens[i].equalsIgnoreCase("-ssl"))
            {
               ssl.assign(true);
            }
            else if (tokens[i].equalsIgnoreCase("-qsize"))
            {
               qsize.assign(Integer.parseInt(tokens[i+1]));
               i++;
            }
            else if (tokens[i].equalsIgnoreCase("-keyalias"))
            {
               keyAlias.assign(tokens[i+1]);
               i++;
            }
            else if (tokens[i].equalsIgnoreCase("-keyaliaspasswd"))
            {
               keyAliasPwd.assign(tokens[i+1]);
               i++;
            }
            else if (tokens[i].equalsIgnoreCase("-nossesioncache "))
            {
               noSessionCache.assign(true);
            }
            else if (tokens[i].equalsIgnoreCase("-sessiontimeout"))
            {
               timeout.assign(Integer.parseInt(tokens[i+1]));
               i++;
            }
            else if (tokens[i].equalsIgnoreCase("-pf"))
            {
               String configFile = tokens[i + 1];
               Stream stream = new StreamWrapper();
               try
               {
                  // the file is on the client
                  stream.assign(StreamFactory.openFileStream(configFile, false, false));
               }
               catch (ErrorConditionException e)
               {
                  // file cannot be accessed
                  errMsg.assign("Unable to open parameter file " + configFile
                  + ", errno 2.");
                  errNo.assign(1247);
                  break;
               }

               // clear any existing errors
               errMsg = new character();

               StringBuffer b = new StringBuffer();
               int asc = stream.readCh();
               // -2 is EoF for FileStreams
               while (asc != -2)
               {
                  // -1 means an error occured while reading
                  if (asc == -1)
                  {
                     errMsg.assign("An error has occured while reading from the parameter file "
                     + configFile + ", errno 2.");
                     break;
                  }
                  char chr = (char) asc;
                  b.append(chr);
                  asc = stream.readCh();
               }
               character fileOpts = new character(b.toString());
               getParams(fileOpts,
                         errMsg,
                         errNo,
                         port,
                         qsize,
                         timeout,
                         keyAlias,
                         keyAliasPwd,
                         ssl,
                         noSessionCache);

            }
         }
      }
   }
   
}
